@@ -5,12 +5,13 @@ Author: Giulio Foletto <giulio.foletto@outlook.com>.
 License: See project-level license file.
 """
 
+from typing import ValuesView
 import networkx as nx
 from . import function_composer
 import inspect
 
 
-starting_node_properties = {"type": "standard", "has_value": False, "value": None, "is_frozen": False, "is_recipe": False}
+starting_node_properties = {"type": "standard", "has_value": False, "value": None, "is_frozen": False, "is_recipe": False, "topological_generation_index": -1, "has_reachability": False, "reachability": None}
 
 
 class Graph():
@@ -54,13 +55,15 @@ class Graph():
             raise ValueError("Cannot add node with dependencies without a recipe")
 
         elif recipe is None:  # Accept nodes with no dependencies
-            # The node is readded even if already existing
-            self._nxdg.add_node(name, **starting_node_properties)
+            # Avoid adding existing node so as not to overwrite attributes
+            if name not in self.nodes:
+                self._nxdg.add_node(name, **starting_node_properties)
 
         else:  # Standard case
             # Add the node
-            # The node is readded even if already existing
-            self._nxdg.add_node(name, **starting_node_properties)
+            # Avoid adding existing node so as not to overwrite attributes
+            if name not in self.nodes:
+                self._nxdg.add_node(name, **starting_node_properties)
             # Set attributes
             # Note: This could be done in the constructor, but doing it separately adds flexibility
             # Indeed, we might want to change how attributes work, and we can do it by modifying setters
@@ -91,37 +94,49 @@ class Graph():
                     self._nxdg.add_node(value, **starting_node_properties)
                 self._nxdg.add_edge(value, name)
 
+    def add_step_quick(self, name, recipe):
+        """
+        Interface to quickly add a step by passing a name and a function.
+
+        The recipe node takes the name of the passed function.
+        Dependency nodes are built from the args and kwonlyargs of the passed function.
+        """
+        # Check that the passed recipe is a valid function
+        if not inspect.isfunction(recipe):
+            raise TypeError("The passed recipe should be a function, but it is a " + type(recipe))
+        argspec = inspect.getfullargspec(recipe)
+        # varargs and varkw are not supported because add_step_quick needs parameter names to build nodes
+        if argspec.varargs is not None or argspec.varkw is not None:
+            raise ValueError("Functions with varargs or varkwargs are not supported by add_step_quick because there would be no way to name dependency nodes")
+
+        # Get function name and parameters
+        recipe_name = recipe.__name__
+        # Lambdas are all automatically named "<lambda>" so we change this
+        if recipe_name == "<lambda>":
+            recipe_name = "recipe_for_" + name
+        args = argspec.args
+        kwargs_list = argspec.kwonlyargs
+        # Build a dictionary with identical keys and values so that recipe is called all the keys are used are kwargs
+        kwargs = {kw: kw for kw in kwargs_list}
+        # Add the step: this will create nodes for name, recipe_name and all elements of args and kwargs_list
+        self.add_step(name, recipe_name, *args, **kwargs)
+        # Directly set the value of recipe_name to recipe
+        self.set_value(recipe_name, recipe)
+
     def add_simple_conditional(self, name, condition, value_true, value_false):
         """
         Interface to add a conditional to the graph.
         """
-        # Add all nodes
-        self._nxdg.add_node(name, **starting_node_properties)
-        for node in [condition, value_true, value_false]:
-            # Avoid adding existing dependencies so as not to overwrite attributes
-            if node not in self.nodes:
-                self._nxdg.add_node(node, **starting_node_properties)
-
-        # Connect edges
-        self._nxdg.add_edge(condition, name)
-        self._nxdg.add_edge(value_true, name)
-        self._nxdg.add_edge(value_false, name)
-
-        # Specify that this node is a conditional
-        self.set_type(name, "conditional")
-
-        # Add conditions name to the list of conditions of the conditional
-        self.set_conditions(name, [condition])
-
-        # Add possibilities to the list of possibilities of the conditional
-        self.set_possibilities(name, [value_true, value_false])
+        self.add_multiple_conditional(name, conditions=[condition], possibilities=[value_true, value_false])
 
     def add_multiple_conditional(self, name, conditions, possibilities):
         """
         Interface to add a multiple conditional to the graph.
         """
         # Add all nodes and connect all edges
-        self._nxdg.add_node(name, **starting_node_properties)
+        # Avoid adding existing node so as not to overwrite attributes
+        if name not in self.nodes:
+            self._nxdg.add_node(name, **starting_node_properties)
         for node in conditions + possibilities:
             # Avoid adding existing dependencies so as not to overwrite attributes
             if node not in self.nodes:
@@ -178,7 +193,7 @@ class Graph():
         if attribute in attributes and attributes[attribute] is not None:
             return attributes[attribute]
         else:
-            raise ValueError("Node ", node, " has no ", attribute)
+            raise ValueError("Node " + node + " has no " + attribute)
 
     def set_node_attribute(self, node, attribute, value):
         self.nodes[node][attribute] = value
@@ -208,15 +223,25 @@ class Graph():
         return self.set_node_attribute(node, "kwargs", kwargs)
 
     def get_conditions(self, node):
-        return self.get_node_attribute(node, "conditions")
+        conditions = self.get_node_attribute(node, "conditions")
+        if not isinstance(conditions, list):
+            conditions = list(conditions)
+        return conditions
 
     def set_conditions(self, node, conditions):
+        if not isinstance(conditions, list):
+            conditions = list(conditions)
         return self.set_node_attribute(node, "conditions", conditions)
 
     def get_possibilities(self, node):
-        return self.get_node_attribute(node, "possibilities")
+        possibilities = self.get_node_attribute(node, "possibilities")
+        if not isinstance(possibilities, list):
+            possibilities = list(possibilities)
+        return possibilities
 
     def set_possibilities(self, node, possibilities):
+        if not isinstance(possibilities, list):
+            possibilities = list(possibilities)
         return self.set_node_attribute(node, "possibilities", possibilities)
 
     def get_type(self, node):
@@ -225,20 +250,44 @@ class Graph():
     def set_type(self, node, type):
         return self.set_node_attribute(node, "type", type)
 
+    def get_topological_generation_index(self, node):
+        return self.get_node_attribute(node, "topological_generation_index")
+
+    def set_topological_generation_index(self, node, index):
+        self.set_node_attribute(node, "topological_generation_index", index)
+
     def get_value(self, node):
         attributes = self.nodes[node]
         if "value" in attributes and attributes["value"] is not None and self.nodes[node]["has_value"]:
             return attributes["value"]
         else:
-            raise ValueError("Node ", node, " has no value")
+            raise ValueError("Node " + node + " has no value")
 
     def set_value(self, node, value):
+        # Note: This changes reachability
         self.nodes[node]["value"] = value
         self.nodes[node]["has_value"] = True
 
     def unset_value(self, node):
+        # Note: This changes reachability
         self.nodes[node]["has_value"] = False
 
+    def get_reachability(self, node):
+        attributes = self.nodes[node]
+        if "reachability" in attributes and attributes["reachability"] is not None and self.nodes[node]["has_reachability"]:
+            return attributes["reachability"]
+        else:
+            raise ValueError("Node " + node + " has no reachability")
+
+    def set_reachability(self, node, reachability):
+        if reachability not in ("unreachable", "uncertain", "reachable"):
+            raise ValueError(reachability + " is not a valid reachability value.")
+        self.nodes[node]["reachability"] = reachability
+        self.nodes[node]["has_reachability"] = True
+
+    def unset_reachability(self, node):
+        self.nodes[node]["has_reachability"] = False
+
     def is_frozen(self, node):
         return self.get_node_attribute(node, "is_frozen")
 
@@ -265,6 +314,26 @@ class Graph():
                 continue
             self.unset_value(node)
 
+    def has_reachability(self, node):
+        return self.get_node_attribute(node, "has_reachability")
+
+    def set_has_reachability(self, node, has_reachability):
+        return self.set_node_attribute(node, "has_reachability", has_reachability)
+
+    def clear_reachabilities(self, *args):
+        """
+        Clear reachabilities in the graph nodes.
+        """
+        if len(args) == 0:  # Interpret as "Clear everything"
+            nodes_to_clear = self.nodes
+        else:
+            nodes_to_clear = args & self.nodes  # Intersection
+
+        for node in nodes_to_clear:
+            if self.is_frozen(node):
+                continue
+            self.unset_reachability(node)
+
     def update_internal_context(self, dictionary):
         """
         Update internal context with a dictionary.
@@ -365,7 +434,7 @@ class Graph():
         elif self.get_type(target) == "conditional":
             return self.evaluate_conditional(target, continue_on_fail)
         else:
-            raise ValueError("Evaluation of nodes of type ", self.get_type(target), " is not supported")
+            raise ValueError("Evaluation of nodes of type " + self.get_type(target) + " is not supported")
 
     def evaluate_standard(self, node, continue_on_fail=False):
         """
@@ -403,20 +472,26 @@ class Graph():
         if self.has_value(conditional):
             self.get_value(conditional)
             return
-        # If not, evaluate the conditions until one is found true
+        # If not, check if one of the conditions already has a true value
         for index, condition in enumerate(self.get_conditions(conditional)):
-            self.evaluate_target(condition, continue_on_fail)
             if self.has_value(condition) and self.get_value(condition):
                 break
-            elif not self.has_value(condition):
-                # Computing failed
-                if continue_on_fail:
-                    # Do nothing, we want to keep going
-                    return
-                else:
-                    raise ValueError("Node " + condition + " could not be computed.")
-        else:  # Happens if loop is never broken, i.e. when no conditions are true
-            index = -1
+        else:
+            # Happens only if loop is never broken
+            # In this case, evaluate the conditions until one is found true
+            for index, condition in enumerate(self.get_conditions(conditional)):
+                self.evaluate_target(condition, continue_on_fail)
+                if self.has_value(condition) and self.get_value(condition):
+                    break
+                elif not self.has_value(condition):
+                    # Computing failed
+                    if continue_on_fail:
+                        # Do nothing, we want to keep going
+                        return
+                    else:
+                        raise ValueError("Node " + condition + " could not be computed")
+            else:  # Happens if loop is never broken, i.e. when no conditions are true
+                index = -1
 
         # Actual computation happens here
         try:
@@ -428,7 +503,7 @@ class Graph():
                 # Do nothing, we want to keep going
                 return
             else:
-                raise ValueError("Node " + possibility + " could not be computed.")
+                raise ValueError("Node " + possibility + " could not be computed")
         # Save results and release
         self.set_value(conditional, res)
 
@@ -446,6 +521,125 @@ class Graph():
         for target in targets:
             self.evaluate_target(target, True)
 
+    def execute_towards_conditions(self, *conditions):
+        """
+        Move towards the conditions, stop if one is found true.
+        """
+        for condition in conditions:
+            self.evaluate_target(condition, True)
+            if self.has_value(condition) and self[condition]:
+                break
+
+    def execute_towards_all_conditions_of_conditional(self, conditional):
+        """
+        Move towards the conditions of a specific conditional, stop if one is found true.
+        """
+        self.execute_towards_conditions(*self.get_conditions(conditional))
+
+    def find_reachability_target(self, target):
+        """
+        Generic interface to find the reachability of a GenericNode.
+        """
+        if self.get_type(target) == "standard":
+            return self.find_reachability_standard(target)
+        elif self.get_type(target) == "conditional":
+            return self.find_reachability_conditional(target)
+        else:
+            raise ValueError("Finding the reachability of nodes of type " + self.get_type(target) + " is not supported")
+
+    def find_reachability_standard(self, node):
+        """
+        Find the reachability of a standard node.
+        """
+        # Check if it already has a reachability
+        if self.has_reachability(node):
+            return
+        # Check if it already has a value
+        if self.has_value(node):
+            self.set_reachability(node, "reachable")
+            return
+        # If not, check the missing dependencies of all arguments
+        dependencies = set(self._nxdg.predecessors(node))
+        if len(dependencies) == 0:
+            # If this node does not have predecessors (and does not have a value itself), it is not reachable
+            self.set_reachability(node, "unreachable")
+            return
+        # Otherwise, dependencies must be checked
+        self.find_reachability_targets(*dependencies)
+        self.set_reachability(node, self.get_worst_reachability(*dependencies))
+
+    def find_reachability_conditional(self, conditional):
+        """
+        Find the reachability of a conditional.
+        """
+        # Check if it already has a reachability
+        if self.has_reachability(conditional):
+            return
+        # Check if it already has a value
+        if self.has_value(conditional):
+            self.get_value(conditional)
+            self.set_reachability(conditional, "reachable")
+            return
+        # If not, evaluate the conditions until one is found true
+        for index, condition in enumerate(self.get_conditions(conditional)):
+            if self.has_value(condition) and self.get_value(condition):
+                # A condition is true
+                possibility = self.get_possibilities(conditional)[index]
+                self.find_reachability_target(possibility)
+                self.set_reachability(conditional, self.get_reachability(possibility))
+                return
+        else:
+            # Happens if loop is never broken, i.e. when no conditions are true
+            # If all conditions and possibilities are reachable -> reachable
+            # If all conditions and possibilities are unreachable -> unreachable
+            # If some conditions are reachable or uncertain but the corresponding possibilities are all unreachable -> unreachable
+            # In all other cases -> uncertain
+            self.find_reachability_targets(*self.get_conditions(conditional))
+            self.find_reachability_targets(*self.get_possibilities(conditional))
+
+            if self.get_worst_reachability(*(self.get_conditions(conditional)+self.get_possibilities(conditional))) == "reachable":
+                # All conditions and possibilities are reachable -> reachable
+                self.set_reachability(conditional, "reachable")
+            elif self.get_best_reachability(*(self.get_conditions(conditional)+self.get_possibilities(conditional))) == "unreachable":
+                # All conditions and possibilities are unreachable -> unreachable
+                self.set_reachability(conditional, "unreachable")
+            else:
+                not_unreachable_condition_possibilities = []
+                for index, condition in enumerate(self.get_conditions(conditional)):
+                    if self.get_reachability(condition) != "unreachable":
+                        not_unreachable_condition_possibilities.append(self.get_possibilities(conditional)[index])
+                if self.get_best_reachability(*not_unreachable_condition_possibilities) == "unreachable":
+                    # All corresponding possibilities are unreachable -> unreachable
+                    self.set_reachability(conditional, "unreachable")
+                else:
+                    self.set_reachability(conditional, "uncertain")
+
+    def get_worst_reachability(self, *nodes):
+        list_of_reachabilities = []
+        for node in nodes:
+            list_of_reachabilities.append(self.get_reachability(node))
+        if "unreachable" in list_of_reachabilities:
+            return "unreachable"
+        elif "uncertain" in list_of_reachabilities:
+            return "uncertain"
+        else:
+            return "reachable"
+
+    def get_best_reachability(self, *nodes):
+        list_of_reachabilities = []
+        for node in nodes:
+            list_of_reachabilities.append(self.get_reachability(node))
+        if "reachable" in list_of_reachabilities:
+            return "reachable"
+        elif "uncertain" in list_of_reachabilities:
+            return "uncertain"
+        else:
+            return "unreachable"
+
+    def find_reachability_targets(self, *targets):
+        for target in targets:
+            self.find_reachability_target(target)
+
     def is_other_node_compatible(self, node, other, other_node):
         # If types differ, return False
         if self.get_type(node) != other.get_type(other_node):
@@ -500,8 +694,10 @@ class Graph():
         subfuncs_dependencies = []
         for argument in self.get_kwargs(node_name):
             if argument == dependency_name:
-                if self.get_type(dependency_name) == "conditional" or self.get_type(self.get_recipe(dependency_name)) == "conditional":
-                    raise TypeError("Simplification does not support conditional nodes.")
+                if self.get_type(dependency_name) != "standard":
+                    raise TypeError("Simplification only supports standard nodes, while the type of " + dependency_name + " is " + self.get_type(dependency_name))
+                if self.get_type(self.get_recipe(dependency_name)) != "standard":
+                    raise TypeError("Simplification only supports standard nodes, while the type of " + self.get_recipe(dependency_name) + " is " + self.get_type(self.get_recipe(dependency_name)))
                 subfuncs.append(self[self.get_recipe(dependency_name)])  # Get python function
                 subfuncs_dependencies.append(list(self.get_args(dependency_name)) + list(self.get_kwargs(dependency_name).values()))
             else:
@@ -523,7 +719,7 @@ class Graph():
     def simplify_all_dependencies(self, node_name, exclude=[]):
         dependencies = self.get_args(node_name) + tuple(self.get_kwargs(node_name).values())
         for dependency in dependencies:
-            if dependency not in exclude and self.get_type(dependency) == "standard":
+            if dependency not in exclude:
                 self.simplify_dependency(node_name, dependency)
 
     def freeze(self, *args):
@@ -563,6 +759,7 @@ class Graph():
         It also marks dependencies of recipes as recipes themselves.
         """
         self.make_recipe_dependencies_also_recipes()
+        self.update_topological_generation_indexes()
         self.freeze()
 
     def get_topological_order(self):
@@ -570,3 +767,109 @@ class Graph():
         Return list of nodes in topological order, i.e., from dependencies to targets
         """
         return list(nx.topological_sort(self._nxdg))
+
+    def get_topological_generations(self):
+        """
+        Return list of topological generations of the graph
+        """
+        return list(nx.topological_generations(self._nxdg))
+
+    def update_topological_generation_indexes(self):
+        generations = self.get_topological_generations()
+        for node in self.nodes:
+            for index, generation in enumerate(generations):
+                if node in generation:
+                    self.set_topological_generation_index(node, index)
+                    break
+
+    def get_all_sources(self, exclude_recipes=False):
+        sources = set()
+        for node in self.nodes:
+            if exclude_recipes and self.is_recipe(node):
+                continue
+            if self._nxdg.in_degree(node) == 0:
+                sources.add(node)
+        return sources
+
+    def get_all_sinks(self, exclude_recipes=False):
+        sinks = set()
+        for node in self.nodes:
+            if exclude_recipes and self.is_recipe(node):
+                continue
+            if self._nxdg.out_degree(node) == 0:
+                sinks.add(node)
+        return sinks
+
+    def convert_conditional_to_trivial_step(self, conditional, execute_towards_conditions=False):
+        """
+        Convert a conditional to a trivial step that returns the dependency corresponding to the true condition.
+
+        Parameters
+        ----------
+        conditional: hashable (typically string)
+            The name of the conditional node to be converted
+        execute_towards_conditions: bool
+            Whether to execute the graph towards the conditions until one is found true (default: False)
+        """
+        if execute_towards_conditions:
+            self.execute_towards_all_conditions_of_conditional(conditional)
+
+        for index, condition in enumerate(self.get_conditions(conditional)):
+            if self.has_value(condition) and self.get_value(condition):
+                break
+        else:  # Happens if loop is never broken, i.e. when no conditions are true
+            if len(self.get_conditions(conditional)) == len(self.get_possibilities(conditional)) - 1:
+                # We assume that the last possibility is considered a default
+                index = -1
+            else:
+                raise ValueError("Cannot convert conditional " + conditional + " if no condition is true")
+        # Get the correct possibility
+        selected_possibility = self.get_possibilities(conditional)[index]
+
+        # Remove all previous edges (the correct one will be readded later)
+        for condition in self.get_conditions(conditional):
+            self._nxdg.remove_edge(condition, conditional)
+        for possibility in self.get_possibilities(conditional):
+            self._nxdg.remove_edge(possibility, conditional)
+        # Rewrite node attributes
+        nx.set_node_attributes(self._nxdg, {conditional: starting_node_properties})
+        # Add a trivial recipe
+        recipe = "trivial_recipe_for_" + conditional
+        self.set_recipe(conditional, recipe)
+        self.set_args(conditional, [selected_possibility])
+        self.set_kwargs(conditional, dict())
+
+        # Add and connect the recipe
+        # Avoid adding existing recipe so as not to overwrite attributes
+        if recipe not in self.nodes:
+            self._nxdg.add_node(recipe, **starting_node_properties)
+        self.set_is_recipe(recipe, True)
+        self._nxdg.add_edge(recipe, conditional)
+        # Assign value of identity function to recipe
+        self.set_value(recipe, lambda x: x)
+
+        # Add and connect the possibility
+        self._nxdg.add_edge(selected_possibility, conditional)
+
+    def get_all_conditionals(self):
+        """
+        Get set of all conditional nodes in the graph.
+        """
+        conditionals = set()
+        for node in self.nodes:
+            if self.get_type(node) == "conditional":
+                conditionals.add(node)
+        return conditionals
+
+    def convert_all_conditionals_to_trivial_steps(self, execute_towards_conditions=False):
+        """
+        Convert all conditionals in the graph to trivial steps that return the dependency corresponding to the true condition.
+
+        Parameters
+        ----------
+        execute_towards_conditions: bool
+            Whether to execute the graph towards the conditions until one is found true (default: False)
+        """
+        conditionals = self.get_all_conditionals()
+        for conditional in conditionals:
+            self.convert_conditional_to_trivial_step(conditional, execute_towards_conditions)
